/*
 *
 * Copyright (c) 2003-2004 Erez Zadok
 * Copyright (c) 2003-2004 Swapnil V. Patil
 * Copyright (c) 2003-2004 Anand Kashyap
 * Copyright (c) 2003-2004 Gopalan Sivathanu
 * Copyright (c) 2003-2004 Stony Brook University
 * Copyright (c) 2003-2004 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */

/*
 *  $Id: i3fs.h,v 1.25 2004/08/18 19:10:49 gopalan Exp $
 */

#ifndef __CACHEFS_H_
#define __CACHEFS_H_

# define ADD_POLICY  _IOR(0x15, 5, int)
# define REMOVE_POLICY _IOR(0x15, 6, int)
# define ALLOW_UPDATES _IOR(0x15, 7, int)
# define DISALLOW_UPDATES _IOR(0x15, 8, int)

#ifdef __KERNEL__

#define SIG_MAX_LEN 20
#define DB_SIZE 100
# define CACHEFS_SUPER_MAGIC 0xf15f

# include "db.h"
# include "i3fs_utils.h"

# define SETUP _IOR(0x15, 4, int)

# define CACHE_SIZE (10 * sizeof(int) * 8) - 1

struct db_node{
	unsigned long ino;
	char checksum[22];
};

struct db_node kernel_db[DB_SIZE];


/*
 * STRUCTURES:
 */


/* Turn this on to include code that counts writes and writes in the middle */
/* #undef FIST_COUNT_WRITES */

/* fist file systems superblock magic */


/* i3fs inode data in memory */
struct i3fs_inode_info {
    int policy;
    int freq_number;
    int check_result;
    int whole_file_result;
    int check_required;
    unsigned long int page_bitmap[10];
    inode_t *wii_inode;
};

/* i3fs dentry data in memory */
struct i3fs_dentry_info {
    dentry_t *wdi_dentry;
};


/* i3fs super-block data in memory */
struct i3fs_sb_info {
    super_block_t *wsi_sb;
    DB *policydb;
    DB *datadb;
    DB *metadb;
    DB *accessdb;
    int allow_updates;
    int cache_results;
    char *passphrase;
    char *mntdir;
    char *logfile;
    int plen;
    struct vfsmount *hidden_mnt;
    ;
};

/* file private data. */
struct i3fs_file_info {
    struct file *wfi_file;
};

/*
 * MACROS:
 */

// File TO Private Data
# define ftopd(file) ((struct i3fs_file_info *)((file)->private_data))
// File TO Hidden File
# define ftohf(file) ((ftopd(file))->wfi_file)
// Inode TO Private Data
# define itopd(ino) ((struct i3fs_inode_info *)(ino)->u.generic_ip)
// Inode TO Hidden Inode
# define itohi(ino) (itopd(ino)->wii_inode)
# define vnode2lower itohi
// Superblock TO Private Data
# define stopd(super) ((struct i3fs_sb_info *)(super)->u.generic_sbp)
# define vfs2priv stopd
// Superblock TO Hidden Superblock
# define stohs(super) (stopd(super)->wsi_sb)
// Dentry TO Hidden Dentry (and private data --- for the future)
# define dtopd(dentry) ((struct i3fs_dentry_info *)(dentry)->d_fsdata)
# define dtohd(dent) (dtopd(dent)->wdi_dentry)
# define dir2lower dtohd
/* VMArea TO Hidden VMArea */
# define vmatohvma(vma) ((vm_area_t *)((vma)->vm_private_data))

# define sbt(sb) ((sb)->s_type->name)

/* macros to simplify non-SCA code */
#  define MALLOC_PAGE_POINTERS(hidden_pages, num_hidden_pages)
#  define MALLOC_PAGEDATA_POINTERS(hidden_pages_data, num_hidden_pages)
#  define FREE_PAGE_POINTERS(hidden_pages, num)
#  define FREE_PAGEDATA_POINTERS(hidden_pages_data, num)
#  define FOR_EACH_PAGE
#  define CURRENT_HIDDEN_PAGE hidden_page
#  define CURRENT_HIDDEN_PAGEDATA hidden_page_data
#  define CURRENT_HIDDEN_PAGEINDEX page->index


/*
 * Flags for i3fs_{en,de}code_filename
 * DO_DOTS means the special entries . and .. should be encoded (for symlink)
 * SKIP_DOTS means they should be preserved intact
 */
#  define DO_DOTS   0
#  define SKIP_DOTS 1


void printvoid(void*, int, int);
int verify_passphrase(DB*, void*);
int remove_policy(char*);
void update_data_db(DB*, unsigned long, unsigned long, void*);
void compute_cksum(void*, int, void*, int, void*);
void insert_policy_in_database(inode_t *, dentry_t *, int, int);
int checksum_metadata(inode_t *, int);
int remove_from_db(DB* , void* , int);
int checksum_filedata_to_db(dentry_t *, file_t *);
void report_metadata_mismatch(struct inode*, struct task_struct*);
void report_data_mismatch(struct dentry *, struct task_struct*);


/*
 * EXTERNALS:
 */
extern struct file_operations i3fs_main_fops;
extern struct file_operations i3fs_dir_fops;
extern struct inode_operations i3fs_main_iops;
extern struct inode_operations i3fs_dir_iops;
extern struct inode_operations i3fs_symlink_iops;
extern struct super_operations i3fs_sops;
extern struct dentry_operations i3fs_dops;
extern struct vm_operations_struct i3fs_shared_vmops;
extern struct vm_operations_struct i3fs_private_vmops;
extern struct address_space_operations i3fs_aops;

extern int i3fs_interpose(dentry_t *hidden_dentry, dentry_t *this_dentry, super_block_t *sb, int flag);
#if defined(FIST_FILTER_DATA) || defined(FIST_FILTER_SCA)
extern page_t *i3fs_get1page(file_t *file, int index);
extern int i3fs_fill_zeros(file_t *file, page_t *page, unsigned from);
# endif /* FIST_FILTER_DATA || FIST_FILTER_SCA */
extern int i3fs_decode_block(const char *from, char *to, int len, const vnode_t *vp, const vfs_t *vfsp, u_long pagenum);
extern int i3fs_encode_block(const char *from, char *to, int len, const vnode_t *vp, const vfs_t *vfsp, u_long pagenum);
extern int i3fs_decode_filename(const char *name, int length, char **decoded_name, int skip_dots, const vnode_t *vp, const vfs_t *vfsp);
extern int i3fs_encode_filename(const char *name, int length, char **encoded_name, int skip_dots, const vnode_t *vp, const vfs_t *vfsp);
extern ino_t i3fs_iunique(struct super_block *sb, ino_t maxreserved);

#  define i3fs_hidden_dentry(d) __i3fs_hidden_dentry(__FILE__,__FUNCTION__,__LINE__,(d))
extern dentry_t *__i3fs_hidden_dentry(char *file, char *func, int line, dentry_t *this_dentry);

int policy_exists(DB*, long int, int*, int*);


#  define copy_inode_size(dst, src) \
    dst->i_size = src->i_size; \
    dst->i_blocks = src->i_blocks;

static inline void
fist_copy_attr_atime(inode_t *dest, const inode_t *src)
{
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_atime = src->i_atime;
}
static inline void
fist_copy_attr_times(inode_t *dest, const inode_t *src)
{
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_atime = src->i_atime;
    dest->i_mtime = src->i_mtime;
    dest->i_ctime = src->i_ctime;
}
static inline void
fist_copy_attr_timesizes(inode_t *dest, const inode_t *src)
{
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_atime = src->i_atime;
    dest->i_mtime = src->i_mtime;
    dest->i_ctime = src->i_ctime;
    copy_inode_size(dest, src);
}
static inline void
fist_copy_attr_all(inode_t *dest, const inode_t *src)
{
    print_entry_location();
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_mode = src->i_mode;
    dest->i_nlink = src->i_nlink;
    dest->i_uid = src->i_uid;
    dest->i_gid = src->i_gid;
    dest->i_rdev = src->i_rdev;
    dest->i_atime = src->i_atime;
    dest->i_mtime = src->i_mtime;
    dest->i_ctime = src->i_ctime;
    dest->i_blksize = src->i_blksize;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,12)
    dest->i_blkbits = src->i_blkbits;
# endif /* linux 2.4.12 and newer */
    copy_inode_size(dest, src);

    dest->i_attr_flags = src->i_attr_flags;
    print_exit_location();
}

/* auto-generated extern definitions */


#endif /* __KERNEL__ */

/*
 * Definitions for user and kernel code
 */

/* ioctls */


#endif	/* not __CACHEFS_H_ */

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
