
#ifndef __I3FS_H_
#define __I3FS_H_

struct policy   {
        char file[1024];
        int policy_bits;
	int freq_number;
	char passphrase[128];
};


#define NO_BLOCK 0

/* Bitmask to represent the policy */
#define PER_PAGE			(1 << 16)
#define INHERIT				(1 << 15)
#define IS_DIR				(1 << 14)
#define PERMS 				(1 << 13)
#define INO 				(1 << 12)
#define NUM_LINKS			(1 << 11)
#define UID				(1 << 10)
#define GID				(1 << 9)
#define TYPE				(1 << 8)
#define SIZE				(1 << 7)
#define DEV_ID				(1 << 6)
#define NUM_BLOCKS			(1 << 5)
#define A_TIME				(1 << 4)
#define M_TIME				(1 << 3)
#define C_TIME				(1 << 2)
#define BLOCK				(1 << 1)
#define DATA				(1 << 0)

/* For Logging */
#define NB_FAIL 0
#define B_SUCCESS 1
#define B_FAILURE 2 


#ifdef __KERNEL__

#include "fist.h"
#include "db.h"

#define I3FS_CACHE_SIZE 10000				// In-kernel DB size ~ 1 MB
#define I3FS_NEGATIVE_CACHE_SIZE 1000		// Negative Cache size ~ 100 KB

#define DATABASE_FILE "my_database_file"	// Name for hidden database file

/* Structure for In-kernel database */
struct i3fs_cache {
	/* Name of file */
	char *name;

	/* Inode Number */
	unsigned long ino;

	/* bitmap policy rules */
	//char[2] policy_bits;
	int policy_bits;

	/* checksum value for the object */
	char *checksum_value;
	
	struct list_head db_cache;
};




/* Structure for Negative Cache */
struct neg_cache {
	/* Name of file */
	char *name;

	/* Inode Number */
	unsigned long ino;

	struct list_head _cache;
};

struct list_head i3fs_db;					// In-kernel Policy DB
static int list_length;

struct list_head negative_cache;			// Negative Cache
static int negative_list_length;

#define CHECKSUM_SIZE 56
dentry_t *root_dentry;
char *logfile;
char *policyfile;

/* Used for readdir */
struct icubefs_getdents_callback {
    void *dirent;
    dentry_t *dentry;
    filldir_t filldir;
    int err;
    int filldir_called;
    int entries_written;
	char *policy_string;
	int policy_len;
	file_t *write_fp;
	char *file_name;
};


struct meta_data	{
	unsigned long           ino;
	kdev_t                  dev_id;
	umode_t                 perm;
	nlink_t                 num_links;
	uid_t                   uid;
	gid_t                   gid;
	loff_t                  size;
	time_t                  atime;
	time_t                  mtime;
	time_t                  ctime;
	unsigned long           num_blocks;
};

int get_checksum(file_t *, char *, int ,char *, int , int );
int get_from_metadb(DB *, long int, void *);
int handle_setup_ioctl(struct super_block *, char *);
unsigned long atol(char *str);
void put_policy_in_database(struct super_block*, struct policy*); 
void convert_policy_to_bitmap(char *, int *, int);
void compute_filedata_checksum(struct file*, super_block_t*, char*);
int get_stored_data_checksum(DB* datadb, long int ino, unsigned long pno, void *checksum);
int icubefs_filldir(void *, const char *, int , loff_t , ino_t , unsigned int );

int frequency_pass(struct inode *);

int md5_initctx(void**);
void md5_freectx(void*);
void md5_append(void*, const unsigned char *, int);
void md5_finish(void* , unsigned char*);

#endif   /* __KERNEL__ */
#endif   /* __I3FS_H__ */
